
/**
 * @imports
 */
import _isString from '@web-native-js/commons/js/isString.js';
import _arrFrom from '@web-native-js/commons/arr/from.js';
import _difference from '@web-native-js/commons/arr/difference.js';
import _with from '@web-native-js/commons/obj/with.js';

import ENV from '../ENV.js';

/**
 * Observes when the given elements or selectors are added or removed
 * from the given context.
 *
 * @param array|HTMLElement|string	els
 * @param function					callback
 * @param object					params
 *
 * @return MutationObserver
 */
export default function(els, callback, params = {}) {
	els = _arrFrom(els, false/*castObject*/);
	var search = (el, nodeListArray) => {
		// Filter out text nodes
		nodeListArray = nodeListArray.filter(node => node.matches);
		if (_isString(el)) {
			// Is directly mutated...
			var matches = nodeListArray.filter(node => node.matches(el));
			// Is contextly mutated...
			if (params.observeIndirectMutation !== false) {
				matches = nodeListArray
					.reduce((collection, node) => collection.concat(_arrFrom(node.querySelectorAll(el))), matches);
				if (matches.length) {
					return matches;
				}
			}
		} else {
			// Is directly mutated...
			if (nodeListArray.includes(el)) {
				return [el];
			}
			// Is contextly mutated...
			if (params.observeIndirectMutation !== false && nodeListArray.length) {
				var parentNode = el;
				while(parentNode = parentNode.parentNode) {
					if (nodeListArray.includes(parentNode)) {
						return [el];
					}
				}
			}
		}
	};
	var connected = [], disconnected = [];
	var subject = params.context || ENV.window.document.documentElement;
	var mo = new ENV.window.MutationObserver(mutations => {
		if (!params.on || params.on === 'connected') {
			var matchedAddedNodes = [];
			els.forEach(el => {
				if (_isString(el)) {
					matchedAddedNodes = mutations
						.reduce((matches, mut) => matches.concat(search(el, _arrFrom(mut.addedNodes)) || []), matchedAddedNodes);
				} else {
					var matchedAsAddedNode = mutations
						.reduce((match, mut) => match || (search(el, _arrFrom(mut.addedNodes)) || [])[0], null);
					if (matchedAsAddedNode) {
						matchedAddedNodes.push(matchedAsAddedNode);
					}
				}
			});
			if (matchedAddedNodes.length) {
				if (params.onceEach) {
					var newlyConnected = _difference(matchedAddedNodes, connected);
					if (newlyConnected.length) {
						connected.push(...newlyConnected);
						callback(newlyConnected, 1);
					}
				} else {
					if (params.once) {
						mo.disconnect();
					}
					callback(matchedAddedNodes, 1);
				}
			}
		}
		if (!params.on || params.on === 'disconnected') {
			var matchedRemovedNodes = [];
			els.forEach(el => {
				if (_isString(el)) {
					matchedRemovedNodes = mutations
						.reduce((matches, mut) => matches.concat(search(el, _arrFrom(mut.removedNodes)) || []), matchedRemovedNodes);
				} else {
					var matchedAsRemovedNode = mutations
						.reduce((match, mut) => match || (search(el, _arrFrom(mut.removedNodes)) || [])[0], null);
					if (matchedAsRemovedNode) {
						matchedRemovedNodes.push(matchedAsRemovedNode);
					}
				}
			});
			if (matchedRemovedNodes.length) {
				if (params.onceEach) {
					var newlyDisconnected = _difference(matchedRemovedNodes, disconnected);
					if (newlyDisconnected.length) {
						disconnected.push(...newlyDisconnected);
						callback(newlyDisconnected, 0);
					}
				} else {
					if (params.once) {
						mo.disconnect();
					}
					callback(matchedRemovedNodes, 0);
				}
			}
		}
	});
	mo.observe(subject, {childList:true, subtree:true});
	return mo;
};