
/**
 * @imports
 */
import _isObject from '@web-native-js/commons/js/isObject.js';
import _any from '@web-native-js/commons/arr/any.js';
import _copy from '@web-native-js/commons/obj/copy.js';
//import Intersection from './Intersection.js';
//import Union from './Union.js';
//import Angle from './Angle.js';

/**
 * The Rect class.
 */
export default class Rect {

    /**
     * Initializes a new Rect instance.
     *
     * @param object	 					props
     *
     * @return void
     */
    constructor(props) {
        Object.keys(props).forEach(key => {
            Object.defineProperty(this, key, {value: props[key]});
        });
    }

    // IntersectionWith(offsetParent, window) + add x/y
    
    /**
     * Recturns the offset values.
     * 
     * @return object
     */
    get offset() {
        return {
            top: this.top,
            left: this.left,
        };
    }

    /**
     * Recturns the size values.
     * 
     * @return object
     */
    get size() {
        return {
            width: this.width,
            height: this.height,
        };
    }

    /**
     * @see Intersection.of().
     *
    * @param Rect|Element|Event|window 	reference
     *
     * @return Intersection
     */
    intersectionWith(reference) {
        return Intersection.of(this, reference);
    }

    /**
     * @see Union.of().
     *
    * @param Rect|Element|Event|window 	reference
     *
     * @return Union
     */
    unionWith(reference) {
        return Union.of(this, reference);
    }
     
    /**
     * @see Angle.of().
     *
    * @param Rect|Element|Event|window 	reference
     *
     * @return Angle
     */
    angleWith(reference) {
        return Angle.of(this, reference);
    }

    /**
     * Gets the element's width/height values or left/top offsets or both.
     *
     * A context can be specified from which to resolve left/top distances.
     * document origins are used by default.
     *
     * This function can calculate the rect of 3 different types of object:
     * - DOM element: offsets are calculated from el.getBoundingClientRect() and resolved relative to the specified params.offsetOrigin.
     *		width and height are calculated from el.getBoundingClientRect().
    * - Event object: offsets are calculated as the event's (client|offset|page) x/y, depending on the specified params.offsetOrigin.
    *		width and height are always 0, 0.
    * - Window object: offsets are calculated as the current left/top scroll, as determined by the value of params.offsetOrigin.
    *		width and height are always the window's inner width/height values.
    *
    * @param Rect|Element|Event|window 	node
    * @param object	 					params
    *
    * @return object
    */
    static of(node, params = {}) {
        if (node instanceof Rect) {
            return node;
        }
        var rectProps = ['width', 'height', 'top', 'right', 'bottom', 'left'];
        var eventProps = ['clientX', 'clientY', 'offsetX', 'offsetY', 'pageX', 'pageY', 'screenX', 'screenY'];
        var _isElement = obj => obj instanceof window.Element || obj instanceof window.HTMLDocument;
        var $props = {};
        if (node !== window/*window also has screenX and screenY*/ 
        && _isObject(node) && _any(eventProps, prop => prop in node)) {
            $props = {width:0, height:0,};
            if (params.offsetOrigin === window || _isElement(params.offsetOrigin)) {
                $props.left = $props.right = node.clientX;
                $props.top = $props.bottom = node.clientY;
            } else if (params.offsetOrigin === true) {
                $props.left = $props.right = node.offsetX;
                $props.top = $props.bottom = node.offsetY;
            } else {
                // Relative to document
                $props.left = $props.right = node.pageX;
                $props.top = $props.bottom = node.pageY;
            }
        } else {
            // For params.offsetOrigin being an element, we'll calculate that at the end of the day
            if (node === window) {
                $props = {width:window.innerWidth, height:window.innerHeight,};
                if (params.offsetOrigin === window 
                    || params.offsetOrigin === true 
                    || _isElement(params.offsetOrigin) /** To be calculated later */) {
                    $props.left = 0;
                    $props.top = 0;
                } else {
                    // Relative to document
                    $props.left = window.pageXOffset;// || document.documentElement.scrollLeft;
                    $props.top = window.pageYOffset;// || document.documentElement.scrollTop;
                }
                $props.right = $props.left + $props.width;
                $props.bottom = $props.top + $props.height;
            } else if (_isElement(node)) {
                $props = node.getBoundingClientRect().toJSON();
                delete $props.x;
                delete $props.y;
                if (params.offsetOrigin === true) {
                    params.offsetOrigin = node.offsetParent; // To be calculated later
                } else if (params.offsetOrigin !== window 
                    && !_isElement(params.offsetOrigin) /** To be calculated later */) {
                    // Relative to document
                    $props.left += window.pageXOffset;// || document.documentElement.scrollLeft);
                    $props.top += window.pageYOffset;// || document.documentElement.scrollTop);
                }
            } else {
                throw new Error('Unknown object type!');
            }
            if (_isElement(params.offsetOrigin)) {
                var contextOffset = Rect.of(params.offsetOrigin, {
                    instantiate: false,
                    offsetOrigin: window,
                });
                $props.left -= contextOffset.left;
                $props.top -= contextOffset.top;
                $props.right -= contextOffset.left;
                $props.bottom -= contextOffset.top;
            }
        }
        $props.node = node;
        if (params.instantiate === false) {
            return $props;
        }
        return new Rect($props);
    }
};